// The Xmas Demo 2021: The Moosery
// Author: Julin
// More info here: https://www.ignorantus.com/pages/xmasdemo2021/
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
#include "moose.h"
#include <math.h>
#include <stdlib.h>

// https://www.wetanz.com/shop/resin-kits/moose-resin-kit

static Scene *s;
static Model *roboscene, *town, *robot, *lower, *upper, *thighL, *thighR, *legL, *legR, *footL, *footR;
static Model *goT, *goL, *goR, *goDst, *steady, *moving, *tri1, *tri2, *laser;

static vec3 motion;

static Material *red, *yellow, *sparkl;

static vec3 xaxis = {1.f,.0f,.0f};
static vec3 yaxis = {.0f,1.f,.0f};
static vec3 zaxis = {.0f,.0f,1.f};

extern Model *txt[3];

Kalman rotK;
Kalman3 posK;
Kalman3 roboK;

static float xpos, ypos=1.75f, zpos, xrot, yrot, zrot, adj1, adj2, adj3, aim_x, kal_x;
static bool ready, second;

static vec3 pos;
static float rot;

static float starttime;
static int startframe;

PID pid[2];
Kalman kal[2];

#define PIF ((float)PI)

typedef struct keyframe_s {
	float timing;
//	int frame;
	vec3 pos;
	float pos_speed;
	float rot;
	float rot_speed;
} KeyFrame;

static KeyFrame keys[] = {
	{  0.f, /*   0,*/ {-3.5f, -0.f,-1.5f}, .1f, PI/3.f, .1f },
	{  6.f, /* 720,*/ {-4.f, -2.f,-4.f}, .001f, PI/4.f, .001f },
	{ 14.f, /* 360,*/ {-12.f, -2.f, -2.f}, .001f, PI/2.f, .001f },
	{ 30.f, /*1800,*/ {-12.f, -2.f,-18.f}, .01f, PI/5.f, .01f },
	{ 33.f, /*2000,*/ {-10.f, -2.f,-18.f}, .01f, PI/5.1f, .1f },
	{ 37.f, /*2200,*/ {-12.f, -2.f,-20.f}, .01f, PI/5.3f, .1f },
	{ 40.f, /*2200,*/ {-13.f, -2.f,-21.f}, .01f, PI/5.f, .1f },
	{ 47.f, /*2400,*/ {-14.f, -100.f,-22.f}, .01f, PI/1.7f, .01f },
	{ 99.f, /*   0,*/ {-14.f, -2.f,-22.f}, 5.1f, PI/4.f, 5.1f },
};

static int key = 0; int nkeys = 9;

void MooseInit(V7X *ctx) {

    s = ctx->scn;

	pid[0] = V7NewPID(.005f, .1f, .005f);
	kal[0] = V7NewKal(.001f, .6f);
	pid[1] = V7NewPID(.005f, .1f, .005f);
	kal[1] = V7NewKal(.001f, .6f);

    red = V7NewMtr(ctx, "Red");
    red->col = (vec4){1.f,.0f,.0f,1.f};
    yellow = V7NewMtr(ctx, "Yellow");
    yellow->col = (vec4){1.f,.1f,.0f,1.f};
    sparkl = V7NewMtr(ctx, "Sparkl");

	roboscene = V7NewMod(ctx, "RoboScene"); // <<<<<<<<<<<<<<<<<<<<<< //
	V7AddMod(s, roboscene);
	roboscene->t = (vec3){-1.f,-2.f,-1.f};

	town = LoadMod("MooseTown");

	robot = LoadMod("SU-76_Rig");
	lower = V7GetMod(ctx, "BodyLower");
	upper = V7GetMod(ctx, "BodyUpper");
	thighR = V7GetMod(ctx, "ThighR");
	thighL = V7GetMod(ctx, "ThighL");
	legR = V7GetMod(ctx, "LegR");
	legL = V7GetMod(ctx, "LegsL");
	footR = V7GetMod(ctx, "FootR");
	footL = V7GetMod(ctx, "FootL");
	laser = V7NewLine(ctx, "Laser", 256); laser->mtr = sparkl;
	goT = V7NewLine(ctx, "GoT", 16);
	goL = V7NewLine(ctx, "GoL", 16);
	goR = V7NewLine(ctx, "GoR", 16);
	goDst = V7NewLine(ctx, "GoDst", 16);

	tri1 = V7NewLine(ctx, "Tri1", 16);
	tri2 = V7NewLine(ctx, "Tri2", 16);
//	V7AddMod(s, tri1);
//	V7Parent(thighR, tri2);
    tri1->mtr = red;
    tri2->mtr = yellow;

	goDst = V7NewLine(ctx, "GoDst", 16);

	V7Parent(roboscene, town);
	V7Parent(roboscene, robot);
	V7Parent(roboscene, goDst);
	V7Parent(roboscene, goT);
	V7Parent(roboscene, goL);
	V7Parent(roboscene, goR);
	V7Parent(upper, laser);

	rotK = V7NewKal(.002f, .62f);
	posK = V7NewKal3(.002f, .62f);
	roboK = V7NewKal3(.002f, .62f);
//#if 0
	V7Line(goT->msh, (vec3){-.2f,.0f,.0f}, (vec3){.2f,.0f,.0f});
	V7Line(goT->msh, (vec3){.0f,-.2f,.0f}, (vec3){.0f,.2f,.0f});
	V7Line(goT->msh, (vec3){.0f,.0f,-.2f}, (vec3){.0f,.0f,.2f});
	V7MeshUp(goT->msh);

	V7Line(goL->msh, (vec3){-.2f,.0f,.0f}, (vec3){.2f,.0f,.0f});
	V7Line(goL->msh, (vec3){.0f,-.2f,.0f}, (vec3){.0f,.2f,.0f});
	V7Line(goL->msh, (vec3){.0f,.0f,-.2f}, (vec3){.0f,.0f,.2f});
	V7MeshUp(goL->msh);

	V7Line(goR->msh, (vec3){-.2f,.0f,.0f}, (vec3){.2f,.0f,.0f});
	V7Line(goR->msh, (vec3){.0f,-.2f,.0f}, (vec3){.0f,.2f,.0f});
	V7Line(goR->msh, (vec3){.0f,.0f,-.2f}, (vec3){.0f,.0f,.2f});
	V7MeshUp(goR->msh);
//#endif

}

void MooseStart(V7X *ctx, int n) {

	second = n;
	steady = goL;
	moving = goR;

	if(!n) {
		starttime = ctx->dat.time;
		goT->t = (vec3){-1.5f,1.7f,.0f};
		goL->t.x = -1.f;
		goR->t.x = -2.f;
		goL->t.y = .33f;
		goR->t.y = .33f;
		goL->t.z = .0f;
		goR->t.z = .33f;
		roboK.E[0] = goT->t.x;
		roboK.E[1] = goT->t.y;
		roboK.E[2] = goT->t.z;
		posK.E[0] = 0.f;
		posK.E[1] = -100.f;
		posK.E[2] = 0.f;
		startframe = ctx->dat.frame;
		motion = (vec3){0.f,0.f,.5f};
		key = 0;
	} else {
		starttime = ctx->dat.time - 30.f;
		goT->t = (vec3){-1.5f,1.7f,20.f};
		goL->t.x = -1.f;
		goR->t.x = -2.f;
		goL->t.y = .33f;
		goR->t.y = .33f;
		goL->t.z = 20.0f;
		goR->t.z = 20.17f;
		roboK.E[0] = goT->t.x;
		roboK.E[1] = goT->t.y;
		roboK.E[2] = goT->t.z;
		posK.E[0] = -30.f;
		posK.E[1] = -2.f;
		posK.E[2] = -50.f;
		rotK.E = 2.f;
		startframe = ctx->dat.frame - 1800;
		motion = (vec3){0.f,0.f,.5f};
		key = 3;
	}

}

void MooseLoop(V7X *ctx) {

	float elapsed = ctx->dat.time - starttime;
	int frame = ctx->dat.frame - startframe;

	V7Parent(roboscene, town);
	V7Parent(roboscene, robot);
	V7Parent(roboscene, goDst);
	V7Parent(roboscene, goT);
	V7Parent(roboscene, goL);
	V7Parent(roboscene, goR);

	if(second && elapsed < 33.f)
		town->dissolve = (33.f - elapsed ) / 3.f;

	int step = second ? 30 : 30;

   	if(ctx->dat.frame%(step*4)==0)
   		aim_x = (rand()%90-45) / 10.f;

   	kal_x = V7Kalman(&kal[0], aim_x);

	robot->t = V7Kalman3(&roboK, &goT->t);
	quat_from_rot(&upper->r, &yaxis, sinf(kal_x)*.5f-(second?.5f:.2f)); // turret rotation

	if(elapsed > keys[key].timing) {
		posK.Q = .002f * keys[key].pos_speed;
		rotK.Q = .002f * keys[key].rot_speed;
	}
	if(elapsed > keys[key+1].timing) {
		key++; key%=nkeys;
	}

	pos = V7Kalman3(&posK, &keys[key].pos);
	pos.x += adj1;
	pos.y += adj2;
	pos.z += adj3;
	roboscene->t = pos; // scene movement

	rot = V7Kalman(&rotK, keys[key].rot);
	rot += yrot;
	quat_from_rot(&roboscene->r, &yaxis, rot); // scene rotation

	if(ctx->dat.frame%step==0) { // switch foot
		Model *tmp=steady; steady=moving; moving=tmp;
		ready=1;
	}

	if(ready) {

		moving->t.y = .35f+.33f*sinf((ctx->dat.frame%step)/((float)step)*PIF); // lift foot
		steady->t.y = .35f; // ankle distance to ground (foot pivot)
		vec3 mov;

		vec3_smul(&mov, 1.f/60.f, &motion);

		vec3_add(&goT->t, &goT->t, &mov);		// move the moose 
		vec3_add(&moving->t, &moving->t, &mov); // moving foot twice
		vec3_add(&moving->t, &moving->t, &mov); // the speed of body
	}

	goT->t.y = (second?1.9f:1.7f)+(second?0.4f:0.1f)*sinf(ctx->dat.time*(step/2.f)); // light bobbing
	if(ctx->dat.frame%step<5) goT->t.y -= .4f; // heavy step

	vec3 A, G; float a, b, c; float r0a, r0b, r1, r2a, r2b, r3; vec2 v2a, v2b;

	float reach = vec3_len(&legR->t) + vec3_len(&footR->t); // assume left leg same length

	V7CalcMtx(lower);

	// LEFT LEG
	trans_from_mat4(&A, &thighL->mat);
	trans_from_mat4(&G, &goL->mat);
	a = vec3_len(&legL->t);
	b = vec3_len(&footL->t);
	c = vec3_dist(&A, &G); //if(c > reach) c = reach; // keep goal within reach
	r1 = PIF-V7AngleFromSides(a, b, c);	// knee angle L
	r0a = -V7AngleFromSides(a, c, b);	// hip angle L(a)
	a = fabsf(A.y - G.y);
	b = c;
	v2a = (vec2){A.x,A.z};
	v2b = (vec2){G.x,G.z};
	c = vec2_dist(&v2a, &v2b);
	r0b = -V7AngleFromSides(a, b, c);	// hip angle L(b)
	quat_from_rot(&thighL->r, &xaxis, r0a+r0b);
	quat_from_rot(&legL->r, &xaxis, r1);
	quat_from_rot(&footL->r, &xaxis, -(r0a+r0b+r1));

	// RIGHT LEG
	trans_from_mat4(&A, &thighR->mat);
	trans_from_mat4(&G, &goR->mat);
	a = vec3_len(&legR->t);
	b = vec3_len(&footR->t);
	c = vec3_dist(&A, &G); //if(c > reach) c = reach; // keep goal within reach
	r3 = PIF-V7AngleFromSides(a, b, c); // knee angle R
	r2a = -V7AngleFromSides(a, c, b); // hip angle R(a)
	a = fabsf(A.y - G.y);
	b = c;
	v2a = (vec2){A.x,A.z};
	v2b = (vec2){G.x,G.z};
	c = vec2_dist(&v2a, &v2b);
	r2b = -V7AngleFromSides(a, b, c); // hip angle R(b)
	quat_from_rot(&thighR->r, &xaxis, r2a+r2b);
	quat_from_rot(&legR->r, &xaxis, r3);
	quat_from_rot(&footR->r, &xaxis, -(r2a+r2b+r3));

	tri1->msh->vvtx.num = 0;
	V7Line(tri1->msh, A, G);

	vec3 ori = (vec3){0,0,0}; vec3 Z = {0,0,1}; Z.z += 1;
	tri2->msh->vvtx.num = 0;
	V7Line(tri2->msh, ori, zaxis);
    V7MeshUp(tri1->msh);
    V7MeshUp(tri2->msh);

	if(ctx->key[_UP_]) xpos+=0.1;
	if(ctx->key[_DOWN_]) xpos-=0.1;
	if(ctx->key[_LEFT_]) zpos-=0.1;
	if(ctx->key[_RIGHT_]) zpos+=0.1;
	if(ctx->key[_PGUP_]) ypos+=0.1;
	if(ctx->key[_PGDN_]) ypos-=0.1;

	if(ctx->key[_Q_]) adj1+=0.1;
	if(ctx->key[_A_]) adj1-=0.1;
	if(ctx->key[_W_]) adj2+=0.1;
	if(ctx->key[_S_]) adj2-=0.1;
	if(ctx->key[_E_]) adj3+=0.1;
	if(ctx->key[_D_]) adj3-=0.1;

	if(ctx->key[_KP2_]) xrot+=0.01;
	if(ctx->key[_KP8_]) xrot-=0.01;
	if(ctx->key[_KP4_]) yrot+=0.01;
	if(ctx->key[_KP6_]) yrot-=0.01;
	if(ctx->key[_KP1_]) zrot+=0.01;
	if(ctx->key[_KP3_]) zrot-=0.01;
#if 0
    if(((int)ctx->dat.time)%5==0)
    for(int i=0; i<10; i++) {
        laser->msh->vvtx.num=10;
        float pos, len, jit;
        pos = rand()/(float)RAND_MAX*10.f;
        len = rand()/(float)RAND_MAX*10.f;
        jit = rand()/(float)RAND_MAX*.05f;
    	V7Line(laser->msh, (vec3){.0f,.77f+jit,0}, (vec3){.0f,.77f+jit,0+len});
        pos = rand()/(float)RAND_MAX*10.f;
        len = rand()/(float)RAND_MAX*10.f;
        jit = rand()/(float)RAND_MAX*.05f;
	    V7Line(laser->msh, (vec3){-.4f,.0f+jit,0}, (vec3){-.4f,.0f+jit,0+len});
    } else laser->msh->vvtx.num=0;
	V7MeshUp(laser->msh);

    sparkl->col = (vec4){rand()/(float)RAND_MAX,rand()/(float)RAND_MAX,rand()/(float)RAND_MAX,.5f};

	quat_from_rot(&roboscene->r, &yaxis, yrot); // scene rotation
#endif
	txt[3]->t.x = -3.f;
	txt[3]->t.y = .0f;
	V7Txt(txt[3],
		"time: %.2f frame: %d key: %.1f %.1f\n"
		"pos: %.2f %.2f %.2f rot: %.2f %.2f %.2f\n"
		"r0: %.2f+%.2f(%.2f)\n"
		"r1:      %.2f(%.2f)\n"
		"r2: %.2f+%.2f(%.2f)\n"
		"r3:      %.2f(%.2f)\n"
		"a%.2f b%.2f c%.2f A-G: %.2f %.2f reach: %.2f adj: %.2f %.2f %.2f",
		elapsed, frame, keys[key].timing, keys[key+1].timing,
		pos.x, pos.y, pos.z, xrot, yrot, zrot,
		r0a, r0b, r0a+r0b, 
		r1, r1,
		r2a, r2b, r2a+r2b,
		r3, r3,
		a, b, c, A.z, G.z, reach, adj1, adj2, adj3);
}

void MooseHide(V7X *ctx) {
	(void)ctx;
	V7Parent(0, town);
	V7Parent(0, robot);
	V7Parent(0, goDst);
	V7Parent(0, goT);
	V7Parent(0, goL);
	V7Parent(0, goR);
	V7Parent(0, tri1);
	V7Parent(0, tri2);
//    V7Hide(roboscene);
//    V7Hide(town);
}
