// The Xmas Demo 2021: Control code for the raytracer.
// Author: Corneliusen
// More info here: https://www.ignorantus.com/pages/xmasdemo2021/
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
#ifndef RAYTRACERH
#define RAYTRACERH

#include <stdint.h>

#include "vec.h"

#define OBJNUM (4*32)

#define MAXW 3840
#define MAXH 2160

#define BLOCKW 16
#define BLOCKH 16

#define MAXWR ((MAXW+BLOCKW-1)&~(BLOCKW-1))
#define MAXHR ((MAXH+BLOCKH-1)&~(BLOCKH-1))

#define BLOCKSX (MAXWR/BLOCKW)
#define BLOCKSY (MAXHR/BLOCKH)

typedef struct {
	v4 obj_pos[OBJNUM];
	v4 obj_col[OBJNUM];
	v4 plane_norm;
	v4 plane_col;
	v4 eye;
	v4 light_pos;
	v4 startpos; // precalc
	float ax; // precalc
	float ay; // precalc
	float pad0;
	float pad1;
	uint8_t probe[BLOCKSX * BLOCKSY];
} gpu_tracer;

#ifndef PI
#define PI 3.1415926535f
#endif

#define SUNPOS 0

#define BLOCK_ANY 0
#define BLOCK_BG  1
#define BLOCK_SPH 2

typedef struct {
	v3 dir;
	float speed;
	v4 obj_col;
	v4 obj_pos;
	float obj_radius;
} sphere;

typedef struct {
	int spherecnt;
	sphere spheres[OBJNUM];
	int frame;
	int curcol;
	int width, height;
} worldinfo;

#define RAD_SPHERE 0.5f

void world_init(  worldinfo *w, gpu_tracer *gt, int width, int height );
void world_timer( worldinfo *w, gpu_tracer *gt );
void world_probe( worldinfo *w, gpu_tracer *gt );

v4 h2rgb( float h );
v4 h2rgb2( float h );
v4 h2rgb3( float h );

#endif
