// Title: Fire Valley
// Based on: Menger's dream
// Original Author: jamiep
// URL: https://www.shadertoy.com/view/lsSSDV
// License: https://creativecommons.org/licenses/by-nc-sa/3.0/
// This modified version was used in the Xmas Demo 2021 by Julin & Corneliusen: https://www.ignorantus.com/pages/xmasdemo2021/

#version 310 es
precision highp float;

//#define XAVIER

in vec2 v_UV1;
uniform float u_Dissolve;

#define v_texCoord v_UV1

layout(location = 0) out vec4 outColor;

layout(std140, binding = 0) uniform ctx_stuff {
    unsigned int width;
    unsigned int height;
    float aspect;
    float time;
    unsigned int frame;
};

layout(std140, binding = 2) uniform stuff
{
	float starttime;
};

#define iTime (time-starttime)

#ifdef XAVIER
#define MaxSteps 20
#define MinimumDistance 0.00005
#define normalDistance     0.001
#define Iterations 7
#define Scale  3.0
#define FieldOfView 3.0
#define NonLinearPerspective 1.0
#define DebugNonlinearPerspective false
#else
#define MaxSteps 40
#define MinimumDistance 0.00005
#define normalDistance     0.001
#define Iterations 7
#define Scale  3.0
#define FieldOfView 3.0
#define NonLinearPerspective 1.0
#define DebugNonlinearPerspective false
#endif

#define Ambient 0.32184
#define Diffuse 0.8
#define LightDir vec3(1.0)
#define LightColor vec3(1.0,0.7,0.358824)
#define LightDir2 vec3(1.0,-1.0,1.0)
#define LightColor2 vec3(0.5,0.633333,0.2)
#define Offset vec3(0.92858,0.92858,0.32858)

#define SLCOLS 7
vec3 slideCol( vec3 v, float h )
{
	vec3 cols[SLCOLS];
	cols[0] = v.rgb;
	cols[1] = v.rgb;
	cols[2] = v.brg;
	cols[3] = v.rgb;
	cols[4] = v.rgb;
	cols[5] = v.gbr;
	cols[6] = cols[0];

	h = h - floor( h );
	float Slice     = float(SLCOLS-1) * h;
	float SliceInt  = floor( Slice );
	float SliceFrac = Slice - SliceInt;

	int i = int(SliceInt);
	return mix( cols[i], cols[i + 1], SliceFrac );
}

vec2 rotate(vec2 v, float a) {
	return vec2(cos(a)*v.x + sin(a)*v.y, -sin(a)*v.x + cos(a)*v.y);
}

// Two light sources. No specular
vec3 getLight(in vec3 color, in vec3 normal, in vec3 dir) {
	vec3 lightDir = normalize(LightDir);
	float diffuse = max(0.0,dot(-normal, lightDir)); // Lambertian

	vec3 lightDir2 = normalize(LightDir2);
	float diffuse2 = max(0.0,dot(-normal, lightDir2)); // Lambertian

	vec3 lc1 = slideCol( LightColor,  iTime/11.0f );
	vec3 lc2 = slideCol( LightColor2, iTime/11.0f );

	return
	(diffuse*Diffuse)*(lc1*color) +
	(diffuse2*Diffuse)*(lc2*color);
}


float DE(in vec3 z)
{
	z  = abs(1.0-mod(z,2.0));
	float time = 0.7*iTime;
	float d = 1.0;
	for (int n = 0; n < Iterations; n++) 
    {
		z.xy = rotate(z.xy,2.0*cos( time/8.0));
		z.yz = rotate(z.yz,2.0*sin( time/200.0));
		z = abs(z);
		if (z.x<z.y){ z.xy = z.yx;}
		if (z.x< z.z){ z.xz = z.zx;}
		if (z.y<z.z){ z.yz = z.zy;}
		z = Scale*z-Offset*(Scale-1.0);
		if( z.z<-0.5*Offset.z*(Scale-1.0))  z.z+=Offset.z*(Scale-1.0);
		d = min(d, length(z) * pow(Scale, float(-n)-1.0));
	}
	return d-normalDistance;
}

// Finite difference normal
vec3 getNormal(in vec3 pos) {
	vec3 e = vec3(0.0,normalDistance,0.0);
	return normalize(vec3(
			DE(pos+e.yxx)-DE(pos-e.yxx),
			DE(pos+e.xyx)-DE(pos-e.xyx),
			DE(pos+e.xxy)-DE(pos-e.xxy)
			)
		);
}

vec3 getColor(vec3 normal, vec3 pos) {
	return vec3(0.7, 0.8, 1.0);
}

vec3 hash3( float n )
{
    return fract(sin(vec3(n,n+1.0,n+2.0))*vec3(43758.5453123,22578.1459123,19642.3490423));
}

vec3 noise( in float x )
{
    float p = floor(x);
    float f = fract(x);
    f = f*f*(3.0-2.0*f);
    return mix( hash3(p+0.0), hash3(p+1.0),f);
}


mat4 rotationMat( in vec3 xyz )
{
    vec3 si = sin(xyz);
    vec3 co = cos(xyz);
	return mat4( co.y*co.z,                co.y*si.z,               -si.y,       0.0,
                 si.x*si.y*co.z-co.x*si.z, si.x*si.y*si.z+co.x*co.z, si.x*co.y,  0.0,
                 co.x*si.y*co.z+si.x*si.z, co.x*si.y*si.z-si.x*co.z, co.x*co.y,  0.0,
			     0.0,                      0.0,                      0.0,        1.0 );
}


vec3 rayMarch(in vec3 from, in vec3 dir) 
{
	float totalDistance = 0.0;
	float distance;
	int steps = 0;
	vec3 pos;
	for (int i=0; i<MaxSteps; i++) 
    {
		pos = from + totalDistance * dir;
		distance = DE(pos);
		totalDistance += distance;
		if (distance < MinimumDistance) break;
		steps = i;
	}
    if (distance>0.1) return exp(2.0*distance)*vec3(1.0, 0.5, 0.0);  
	float smoothStep =   float(steps) + distance/MinimumDistance;
	float ao = 1.1-smoothStep/float(MaxSteps);
	vec3 normal = getNormal(pos-dir*normalDistance*3.0);
	vec3 color = getColor(normal, pos);
	vec3 light = getLight(color, normal, dir);
	color = (color*Ambient+light)*ao;
	return sqrt(color);
}

void main()
{
	vec2 q = v_UV1;
    vec2 p = -1.0 + 2.0 * q;
    p.x *= aspect;
    vec2 m = vec2(0.5);

    // animation	
	float time = iTime;

	float time1 = (time-10.0)*1.5 - 167.0;
	float time2 = time;
    const float s = 1.1;
	mat4 mm;
    mm = rotationMat( vec3(0.4,0.1,3.4) + 
                      0.15*sin(0.1*vec3(0.40,0.30,0.61)*time1) + 
                      0.15*sin(0.1*vec3(0.11,0.53,0.48)*time1));
    mm[0].xyz *= s;	
    mm[1].xyz *= s;
    mm[2].xyz *= s;	
	mm[3].xyz = vec3( 0.15, 0.05, -0.07 ) + 0.05*sin(vec3(0.0,1.0,2.0) + 0.2*vec3(0.31,0.24,0.42)*time1);
	
    // camera
	float an = 1.0 + 0.1*time2 - 6.2*m.x;
	vec3 ro = vec3( 0.0f, 0.0f, sin(time2*0.2f) );
    
    vec3 ta = vec3( 0.0, 0.0 + 0.13*(cos(0.3*time/64.0f)+4.0f), 0.0 );
	ta += 0.05*noise(  0.0 + 1.0 );
    vec3 ww = normalize( ta - ro );
    vec3 uu = normalize( cross(ww,vec3(0.0,1.0,0.0) ) );
    vec3 vv = normalize( cross(uu,ww));
    vec3 rd = normalize( p.x*uu + p.y*vv + 3.0*ww );

	outColor = vec4( rayMarch(ro, rd), 1.0f - u_Dissolve );
}
