// Title: Fractal Noise background effect
// Based on: Fractal Noise
// Original author: mu6k
// URL: https://www.shadertoy.com/view/Msf3Wr
// License: https://creativecommons.org/licenses/by-nc-sa/3.0/
// This modified version was used in the Xmas Demo 2021 by Julin & Corneliusen: https://www.ignorantus.com/pages/xmasdemo2021/

#version 310 es
precision highp float;

//#define XAVIER

in vec2 v_UV1;
uniform float u_Dissolve;

#define v_texCoord v_UV1

layout(location = 0) out vec4 outColor;

layout(std140, binding = 0) uniform ctx_stuff {
    unsigned int width;
    unsigned int height;
    float aspect;
    float time;
    unsigned int frame;
};

#define iTime time

float hash(float x)
{
	return fract(sin(cos(x*12.13)*19.123)*17.321);
}

float noise(vec2 p)
{
	vec2 pm = mod(p,1.0);
	vec2 pd = p-pm;
	float v0=hash(pd.x+pd.y*41.0);
	float v1=hash(pd.x+1.0+pd.y*41.0);
	float v2=hash(pd.x+pd.y*41.0+41.0);
	float v3=hash(pd.x+pd.y*41.0+42.0);
	v0 = mix(v0,v1,smoothstep(0.0,1.0,pm.x));
	v2 = mix(v2,v3,smoothstep(0.0,1.0,pm.x));
	return mix(v0,v2,smoothstep(0.0,1.0,pm.y));
}

void main()
{
	float iTime2 = -iTime;
	vec2 uv = v_UV1 - 0.5f;
	
	float rot=sin(iTime2*0.3)*sin(iTime2*0.4)*0.2;
		
	uv *= mat2(cos(rot),-sin(rot),sin(rot),cos(rot));
    
	float v =0.0;
	
	for (float i = 0.0; i<12.0; i+=1.0)
	{
		float t = mod(iTime2+i,12.0);
		float l = iTime2-t;
		float e = exp2(t);
		v+=noise(uv*e+vec2(cos(l)*53.0,sin(l)*100.0))*(1.0-(t/12.0))*(t/12.0);
	}
	
	v-=0.5;
	
	vec3 color = 1.0f-sqrt(vec3(v));

	outColor = vec4(color,1.0-u_Dissolve);
}
